@@ -1,7 +1,7 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2010-2014 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2010-2015 -- leonerd@leonerd.org.uk
 
 package Tangence::Client;
 
@@ -10,7 +10,7 @@ use warnings;
 
 use base qw( Tangence::Stream );
 
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 
 use Carp;
 
@@ -18,6 +18,8 @@ use Tangence::Constants;
 use Tangence::Types;
 use Tangence::ObjectProxy;
 
+use Future;
+
 use List::Util qw( max );
 
 use constant VERSION_MINOR_MIN => 2;
@@ -89,7 +91,9 @@ The following methods are provided by this mixin.
 # Accessors for Tangence::Message decoupling
 sub objectproxies { shift->{objectproxies} ||= {} }
 
-=head2 $rootobj = $client->rootobj
+=head2 rootobj
+
+   $rootobj = $client->rootobj
 
 Returns a L<Tangence::ObjectProxy> to the server's root object
 
@@ -102,7 +106,9 @@ sub rootobj
    return $self->{rootobj};
 }
 
-=head2 $registry = $client->registry
+=head2 registry
+
+   $registry = $client->registry
 
 Returns a L<Tangence::ObjectProxy> to the server's object registry
 
@@ -122,7 +128,9 @@ sub on_error
    return $self->{on_error};
 }
 
-=head2 $client->tangence_connected( %args )
+=head2 tangence_connected
+
+   $client->tangence_connected( %args )
 
 Once the base connection to the server has been established, this method
 should be called to perform the initial work of requesting the root object and
@@ -317,8 +325,8 @@ sub make_proxy
 
    my $obj = $self->objectproxies->{$id} =
       Tangence::ObjectProxy->new(
-         conn => $self,
-         id   => $id,
+         client => $self,
+         id     => $id,
 
          class => $class,
 
@@ -330,6 +338,25 @@ sub make_proxy
    return $obj;
 }
 
+=head1 SUBCLASSING METHODS
+
+These methods are intended for implementation classes to override.
+
+=cut
+
+=head2 new_future
+
+   $f = $client->new_future
+
+Returns a new L<Future> instance for basing asynchronous operations on.
+
+=cut
+
+sub new_future
+{
+   return Future->new;
+}
+
 =head1 AUTHOR
 
 Paul Evans <leonerd@leonerd.org.uk>
