@@ -1,14 +1,14 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2010-2013 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2010-2015 -- leonerd@leonerd.org.uk
 
 package Tangence::ObjectProxy;
 
 use strict;
 use warnings;
 
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 
 use Carp;
 
@@ -43,8 +43,8 @@ sub new
    my %args = @_;
 
    my $self = bless {
-      conn => $args{conn},
-      id   => $args{id},
+      client => $args{client},
+      id     => $args{id},
 
       class => $args{class},
 
@@ -52,7 +52,7 @@ sub new
    }, $class;
 
    # An ObjectProxy is useless after its connection disappears
-   weaken( $self->{conn} );
+   weaken( $self->{client} );
 
    return $self;
 }
@@ -73,6 +73,9 @@ sub destroy
 
 =head1 METHODS
 
+The following methods documented with a trailing call to C<< ->get >> return
+L<Future> instances.
+
 =cut
 
 use overload '""' => \&STRING;
@@ -83,7 +86,9 @@ sub STRING
    return "Tangence::ObjectProxy[id=$self->{id}]";
 }
 
-=head2 $id = $proxy->id
+=head2 id
+
+   $id = $proxy->id
 
 Returns the object ID for the C<Tangence> object being proxied for.
 
@@ -95,7 +100,9 @@ sub id
    return $self->{id};
 }
 
-=head2 $classname = $proxy->classname
+=head2 classname
+
+   $classname = $proxy->classname
 
 Returns the name of the class of the C<Tangence> object being proxied for.
 
@@ -107,7 +114,9 @@ sub classname
    return $self->{class}->name;
 }
 
-=head2 $class = $proxyobj->class
+=head2 class
+
+   $class = $proxyobj->class
 
 Returns the L<Tangence::Meta::Class> object representing the class of this
 object.
@@ -120,7 +129,9 @@ sub class
    return $self->{class};
 }
 
-=head2 $method = $proxy->can_method( $name )
+=head2 can_method
+
+   $method = $proxy->can_method( $name )
 
 Returns the L<Tangence::Meta::Method> object representing the named method, or
 C<undef> if no such method exists.
@@ -133,7 +144,9 @@ sub can_method
    return $self->class->method( @_ );
 }
 
-=head2 $event = $proxy->can_event( $name )
+=head2 can_event
+
+   $event = $proxy->can_event( $name )
 
 Returns the L<Tangence::Meta::Event> object representing the named event, or
 C<undef> if no such event exists.
@@ -146,7 +159,9 @@ sub can_event
    return $self->class->event( @_ );
 }
 
-=head2 $property = $proxy->can_property( $name )
+=head2 can_property
+
+   $property = $proxy->can_property( $name )
 
 Returns the L<Tangence::Meta::Property> object representing the named
 property, or C<undef> if no such property exists.
@@ -191,124 +206,73 @@ sub grab
    }
 }
 
-=head2 $proxy->call_method( %args )
+=head2 call_method
 
-Calls the given method on the server object and invokes a callback function
-when a result is received.
+   $result = $proxy->call_method( $mname, @args )->get
 
-Takes the following named arguments:
-
-=over 8
-
-=item method => STRING
-
-The name of the method
-
-=item args => ARRAY
-
-Optional. If provided, gives positional arguments for the method.
-
-=item on_result => CODE
-
-Callback function to invoke when a result is returned
-
- $on_result->( $result )
-
-=item on_error => CODE
-
-Optional. Callback function to invoke when an error is returned. The client's
-default will apply if not provided.
-
- $on_error->( $error )
-
-=back
+Calls the given method on the server object, passing in the given arguments.
+Returns a L<Future> that will yield the method's result.
 
 =cut
 
 sub call_method
 {
    my $self = shift;
-   my %args = @_;
+   my ( $method, @args ) = @_;
 
-   my $method = delete $args{method} or croak "Need a method";
-   my $args   = delete $args{args};
-
-   ref( my $on_result = delete $args{on_result} ) eq "CODE" 
-      or croak "Expected 'on_result' as a CODE ref";
-
-   my $on_error = delete $args{on_error} || $self->{on_error};
-   ref $on_error eq "CODE" or croak "Expected 'on_error' as a CODE ref";
+   # Detect void-context legacy uses
+   defined wantarray or
+      croak "->call_method in void context no longer useful - it now returns a Future";
 
    my $mdef = $self->can_method( $method )
       or croak "Class ".$self->classname." does not have a method $method";
 
-   my $conn = $self->{conn};
+   my $client = $self->{client};
 
-   my $request = Tangence::Message->new( $conn, MSG_CALL )
+   my $request = Tangence::Message->new( $client, MSG_CALL )
          ->pack_int( $self->id )
          ->pack_str( $method );
 
    my @argtypes = $mdef->argtypes;
-   $argtypes[$_]->pack_value( $request, $args ? $args->[$_] : undef ) for 0..$#argtypes;
+   $argtypes[$_]->pack_value( $request, $args[$_] ) for 0..$#argtypes;
 
-   $conn->request(
+   $client->request(
       request => $request,
+   )->then( sub {
+      my ( $message ) = @_;
 
-      on_response => sub {
-         my ( $message ) = @_;
-         my $code = $message->code;
+      my $code = $message->code;
 
-         if( $code == MSG_RESULT ) {
-            my $result = $mdef->ret ? $mdef->ret->unpack_value( $message )
-                                    : undef;
-            $on_result->( $result );
-         }
-         elsif( $code == MSG_ERROR ) {
-            my $msg = $message->unpack_str();
-            $on_error->( $msg );
-         }
-         else {
-            $on_error->( "Unexpected response code $code" );
-         }
-      },
-   );
+      if( $code == MSG_RESULT ) {
+         my $result = $mdef->ret ? $mdef->ret->unpack_value( $message )
+                                 : undef;
+         Future->done( $result );
+      }
+      else {
+         Future->fail( "Unexpected response code $code", tangence => );
+      }
+   });
 }
 
-=head2 $proxy->subscribe_event( %args )
+=head2 subscribe_event
+
+   $proxy->subscribe_event( $event, %callbacks )->get
 
 Subscribes to the given event on the server object, installing a callback
 function which will be invoked whenever the event is fired.
 
-Takes the following named arguments:
+Takes the following named callbacks:
 
 =over 8
 
-=item event => STRING
-
-Name of the event
-
 =item on_fire => CODE
 
 Callback function to invoke whenever the event is fired
 
  $on_fire->( @args )
 
-=item on_subscribed => CODE
-
-Optional. Callback function to invoke once the event subscription is
-successfully installed by the server.
-
- $on_subscribed->()
-
-If this is provided, it is guaranteed to be invoked before any invocation of
-the C<on_fire> event handler.
-
-=item on_error => CODE
-
-Optional. Callback function to invoke when an error is returned. The client's
-default will apply if not provided.
-
- $on_error->( $error )
+The returned C<Future> it is guaranteed to be completed before any invocation
+of the C<on_fire> event handler.
 
 =back
 
@@ -317,52 +281,45 @@ default will apply if not provided.
 sub subscribe_event
 {
    my $self = shift;
-   my %args = @_;
+   my ( $event, %args ) = @_;
+
+   # Detect void-context legacy uses
+   defined wantarray or
+      croak "->subscribe_event in void context no longer useful - it now returns a Future";
 
-   my $event = delete $args{event} or croak "Need a event";
    ref( my $callback = delete $args{on_fire} ) eq "CODE"
       or croak "Expected 'on_fire' as a CODE ref";
 
-   my $on_error = delete $args{on_error} || $self->{on_error};
-   ref $on_error eq "CODE" or croak "Expected 'on_error' as a CODE ref";
-
-   my $on_subscribed = $args{on_subscribed};
-
    $self->can_event( $event )
       or croak "Class ".$self->classname." does not have an event $event";
 
    if( my $cbs = $self->{subscriptions}->{$event} ) {
       push @$cbs, $callback;
-      return;
+      return Future->done;
    }
 
    my @cbs = ( $callback );
    $self->{subscriptions}->{$event} = \@cbs;
 
-   return if $event eq "destroy"; # This is automatically handled
+   return Future->done if $event eq "destroy"; # This is automatically handled
 
-   my $conn = $self->{conn};
-   $conn->request(
-      request => Tangence::Message->new( $conn, MSG_SUBSCRIBE )
+   my $client = $self->{client};
+
+   $client->request(
+      request => Tangence::Message->new( $client, MSG_SUBSCRIBE )
          ->pack_int( $self->id )
          ->pack_str( $event ),
+   )->then( sub {
+      my ( $message ) = @_;
+      my $code = $message->code;
 
-      on_response => sub {
-         my ( $message ) = @_;
-         my $code = $message->code;
-
-         if( $code == MSG_SUBSCRIBED ) {
-            $on_subscribed->() if $on_subscribed;
-         }
-         elsif( $code == MSG_ERROR ) {
-            my $msg = $message->unpack_str();
-            $on_error->( $msg );
-         }
-         else {
-            $on_error->( "Unexpected response code $code" );
-         }
-      },
-   );
+      if( $code == MSG_SUBSCRIBED ) {
+         Future->done;
+      }
+      else {
+         Future->fail( "Unexpected response code $code", tangence => );
+      }
+   });
 }
 
 sub handle_request_EVENT
@@ -380,38 +337,28 @@ sub handle_request_EVENT
    }
 }
 
-=head2 $proxy->unsubscribe_event( %args )
+=head2 unsubscribe_event
+
+   $proxy->unsubscribe_event( $event )
 
 Removes an event subscription on the given event on the server object that was
 previously installed using C<subscribe_event>.
 
-Takes the following named arguments:
-
-=over 8
-
-=item event => STRING
-
-Name of the event
-
-=back
-
 =cut
 
 sub unsubscribe_event
 {
    my $self = shift;
-   my %args = @_;
-
-   my $event = delete $args{event} or croak "Need a event";
+   my ( $event ) = @_;
 
    $self->can_event( $event )
       or croak "Class ".$self->classname." does not have an event $event";
 
    return if $event eq "destroy"; # This is automatically handled
 
-   my $conn = $self->{conn};
-   $conn->request(
-      request => Tangence::Message->new( $conn, MSG_UNSUBSCRIBE )
+   my $client = $self->{client};
+   $client->request(
+      request => Tangence::Message->new( $client, MSG_UNSUBSCRIBE )
          ->pack_int( $self->id )
          ->pack_str( $event ),
 
@@ -419,173 +366,102 @@ sub unsubscribe_event
    );
 }
 
-=head2 $proxy->get_property( %args )
+=head2 get_property
 
-Requests the current value of the property from the server object, and invokes
-a callback function when the value is received.
+   $value = $proxy->get_property( $prop )->get
 
-Takes the following named arguments
-
-=over 8
-
-=item property => STRING
-
-The name of the property
-
-=item on_value => CODE
-
-Callback function to invoke when the value is returned
-
- $on_value->( $value )
-
-=item on_error => CODE
-
-Optional. Callback function to invoke when an error is returned. The client's
-default will apply if not provided.
-
- $on_error->( $error )
-
-=back
+Requests the current value of the property from the server object.
 
 =cut
 
 sub get_property
 {
    my $self = shift;
-   my %args = @_;
+   my ( $property ) = @_;
 
-   my $property = delete $args{property} or croak "Need a property";
-
-   ref( my $on_value = delete $args{on_value} ) eq "CODE" 
-      or croak "Expected 'on_value' as a CODE ref";
-
-   my $on_error = delete $args{on_error} || $self->{on_error};
-   ref $on_error eq "CODE" or croak "Expected 'on_error' as a CODE ref";
+   # Detect void-context legacy uses
+   defined wantarray or
+      croak "->get_property in void context no longer useful - it now returns a Future";
 
    my $pdef = $self->can_property( $property )
       or croak "Class ".$self->classname." does not have a property $property";
 
-   my $conn = $self->{conn};
-   $conn->request(
-      request => Tangence::Message->new( $conn, MSG_GETPROP )
+   my $client = $self->{client};
+   $client->request(
+      request => Tangence::Message->new( $client, MSG_GETPROP )
          ->pack_int( $self->id )
          ->pack_str( $property ),
+   )->then( sub {
+      my ( $message ) = @_;
+      my $code = $message->code;
 
-      on_response => sub {
-         my ( $message ) = @_;
-         my $code = $message->code;
-
-         if( $code == MSG_RESULT ) {
-            my $value = $pdef->overall_type->unpack_value( $message );
-            $on_value->( $value );
-         }
-         elsif( $code == MSG_ERROR ) {
-            my $msg = $message->unpack_str();
-            $on_error->( $msg );
-         }
-         else {
-            $on_error->( "Unexpected response code $code" );
-         }
-      },
-   );
+      if( $code == MSG_RESULT ) {
+         my $value = $pdef->overall_type->unpack_value( $message );
+         Future->done( $value );
+      }
+      else {
+         Future->fail( "Unexpected response code $code", tangence => );
+      }
+   });
 }
 
-=head2 $proxy->get_property_element( %args )
+=head2 get_property_element
+
+   $value = $proxy->get_property_element( $property, $index_or_key )->get
 
 Requests the current value of an element of the property from the server
-object, and invokes a callback function when the value is received.
-
-Takes the following named arguments
-
-=over 8
-
-=item property => STRING
-
-The name of the property
-
-=item index => INT
-
-For queue or array dimension properties, the index of the element
-
-=item key => STRING
-
-For hash dimension properties, the key of the element
-
-=item on_value => CODE
-
-Callback function to invoke when the value is returned
-
- $on_value->( $value )
-
-=item on_error => CODE
-
-Optional. Callback function to invoke when an error is returned. The client's
-default will apply if not provided.
-
- $on_error->( $error )
-
-=back
+object.
 
 =cut
 
 sub get_property_element
 {
    my $self = shift;
-   my %args = @_;
+   my ( $property, $index_or_key ) = @_;
 
-   my $property = delete $args{property} or croak "Need a property";
-
-   ref( my $on_value = delete $args{on_value} ) eq "CODE" 
-      or croak "Expected 'on_value' as a CODE ref";
-
-   my $on_error = delete $args{on_error} || $self->{on_error};
-   ref $on_error eq "CODE" or croak "Expected 'on_error' as a CODE ref";
+   # Detect void-context legacy uses
+   defined wantarray or
+      croak "->get_property_element in void context no longer useful - it now returns a Future";
 
    my $pdef = $self->can_property( $property )
       or croak "Class ".$self->classname." does not have a property $property";
 
-   my $conn = $self->{conn};
-   $conn->_ver_can_getpropelem or croak "Server is too old to support MSG_GETPROPELEM";
+   my $client = $self->{client};
+   $client->_ver_can_getpropelem or croak "Server is too old to support MSG_GETPROPELEM";
 
-   my $request = Tangence::Message->new( $conn, MSG_GETPROPELEM )
+   my $request = Tangence::Message->new( $client, MSG_GETPROPELEM )
       ->pack_int( $self->id )
       ->pack_str( $property );
 
    if( $pdef->dimension == DIM_HASH ) {
-      defined $args{key} or croak "Need a key";
-      $request->pack_str( $args{key} );
+      $request->pack_str( $index_or_key );
    }
    elsif( $pdef->dimension == DIM_ARRAY or $pdef->dimension == DIM_QUEUE ) {
-      defined $args{index} or croak "Need an index";
-      $request->pack_int( $args{index} );
+      $request->pack_int( $index_or_key );
    }
    else {
-      croak "Cannot get_property_element of a non hash";
+      croak "Cannot get_property_element of a non hash, array or queue";
    }
 
-   $conn->request(
+   $client->request(
       request => $request,
+   )->then( sub {
+      my ( $message ) = @_;
+      my $code = $message->code;
 
-      on_response => sub {
-         my ( $message ) = @_;
-         my $code = $message->code;
-
-         if( $code == MSG_RESULT ) {
-            my $value = $pdef->type->unpack_value( $message );
-            $on_value->( $value );
-         }
-         elsif( $code == MSG_ERROR ) {
-            my $msg = $message->unpack_str();
-            $on_error->( $msg );
-         }
-         else {
-            $on_error->( "Unexpected response code $code" );
-         }
-      },
-   );
+      if( $code == MSG_RESULT ) {
+         my $value = $pdef->type->unpack_value( $message );
+         Future->done( $value );
+      }
+      else {
+         Future->fail( "Unexpected response code $code", tangence => );
+      }
+   });
 }
 
-=head2 $value = $proxy->prop( $property )
+=head2 prop
+
+   $value = $proxy->prop( $property )
 
 Returns the locally-cached value of a smashed property. If the named property
 is not a smashed property, an exception is thrown.
@@ -604,128 +480,65 @@ sub prop
    croak "$self does not have a cached property '$property'";
 }
 
-=head2 $proxy->set_property( %args )
+=head2 set_property
 
-Sets the value of the property in the server object. Optionally invokes a
-callback function when complete.
+   $proxy->set_property( $prop, $value )->get
 
-Takes the following named arguments
-
-=over 8
-
-=item property => STRING
-
-The name of the property
-
-=item value => SCALAR
-
-New value to set for the property
-
-=item on_done => CODE
-
-Optional. Callback function to invoke once the new value is set.
-
- $on_done->()
-
-=item on_error => CODE
-
-Optional. Callback function to invoke when an error is returned. The client's
-default will apply if not provided.
-
- $on_error->( $error )
-
-=back
+Sets the value of the property in the server object.
 
 =cut
 
 sub set_property
 {
    my $self = shift;
-   my %args = @_;
+   my ( $property, $value ) = @_;
 
-   my $property = delete $args{property} or croak "Need a property";
-
-   my $on_done = delete $args{on_done};
-   !defined $on_done or ref $on_done eq "CODE"
-      or croak "Expected 'on_done' to be a CODE ref";
-
-   my $on_error = delete $args{on_error} || $self->{on_error};
-   ref $on_error eq "CODE" or croak "Expected 'on_error' as a CODE ref";
-
-   # value can quite legitimately be undef
-   exists $args{value} or croak "Need a value";
-   my $value = delete $args{value};
+   # Detect void-context legacy uses
+   defined wantarray or
+      croak "->set_property in void context no longer useful - it now returns a Future";
 
    my $pdef = $self->can_property( $property )
       or croak "Class ".$self->classname." does not have a property $property";
 
-   my $conn = $self->{conn};
-   my $request = Tangence::Message->new( $conn, MSG_SETPROP )
+   my $client = $self->{client};
+   my $request = Tangence::Message->new( $client, MSG_SETPROP )
          ->pack_int( $self->id )
          ->pack_str( $property );
    $pdef->type->pack_value( $request, $value ),
 
-   $conn->request(
+   $client->request(
       request => $request,
+   )->then( sub {
+      my ( $message ) = @_;
+      my $code = $message->code;
 
-      on_response => sub {
-         my ( $message ) = @_;
-         my $code = $message->code;
-
-         if( $code == MSG_OK ) {
-            $on_done->() if $on_done;
-         }
-         elsif( $code == MSG_ERROR ) {
-            my $msg = $message->unpack_str();
-            $on_error->( $msg );
-         }
-         else {
-            $on_error->( "Unexpected response code $code" );
-         }
-      },
-   );
+      if( $code == MSG_OK ) {
+         Future->done;
+      }
+      else {
+         Future->fail( "Unexpected response code $code", tangence => );
+      }
+   });
 }
 
-=head2 $proxy->watch_property( %args )
+=head2 watch_property
+
+   $proxy->watch_property( $property, %callbacks )->get
+
+=head2 watch_property_with_initial
+
+   $proxy->watch_property_with_initial( $property, %callbacks )->get
 
 Watches the given property on the server object, installing callback functions
-which will be invoked whenever the property value changes.
+which will be invoked whenever the property value changes. The latter form
+additionally ensures that the server will send the current value of the
+property as an initial update to the C<on_set> event, atomically when it
+installs the update watches.
 
 Takes the following named arguments:
 
 =over 8
 
-=item property => STRING
-
-Name of the property
-
-=item want_initial => BOOLEAN
-
-Optional. If true, requests that the server send the current value of the
-property at the time the watch is installed, in an C<on_set> event. This is
-performed atomically with installing watch.
-
-=item iter_from => INT
-
-Optional. If defined, requests that the server create an iterator for the
-property value (whose dimension must be a queue). Its value indicates which
-end of the queue the iterator should start from; C<ITER_FIRST> to start at
-index 0, or C<ITER_LAST> to start at the highest-numbered index. The iterator
-object will be returned to the C<on_iter> callback. The iterator is
-constructed atomically with installing the watch.
-
-This option is mutually-exclusive with C<want_initial>.
-
-=item on_watched => CODE
-
-Optional. Callback function to invoke once the property watch is
-successfully installed by the server.
-
- $on_watched->()
-
-If this is provided, it is guaranteed to be invoked before any invocation of
-the value change handlers.
-
 =item on_updated => CODE
 
 Optional. Callback function to invoke whenever the property value changes.
@@ -735,22 +548,6 @@ Optional. Callback function to invoke whenever the property value changes.
 If not provided, then individual handlers for individual change types must be
 provided.
 
-=item on_iter => CODE
-
-Callback function to invoke when the iterator object is returned by the
-server. This must be provided if C<iter_from> is provided. It is passed the
-iterator object, and the first and last indices that the iterator will yield
-(inclusive).
-
- $on_iter->( $iter, $first_idx, $last_idx )
-
-=item on_error => CODE
-
-Optional. Callback function to invoke when an error is returned. The client's
-default will apply if not provided.
-
- $on_error->( $error )
-
 =back
 
 The set of callback functions that are required depends on the type of the
@@ -759,42 +556,9 @@ L<Tangence::Object>.
 
 =cut
 
-sub watch_property
+sub _watchcbs_from_args
 {
-   my $self = shift;
-   my %args = @_;
-
-   my $property = delete $args{property} or croak "Need a property";
-
-   my $on_error = delete $args{on_error} || $self->{on_error};
-   ref $on_error eq "CODE" or croak "Expected 'on_error' as a CODE ref";
-
-   my $want_initial = delete $args{want_initial};
-
-   my $iter_from = delete $args{iter_from};
-   if( !defined $iter_from ) {
-      # ignore
-   }
-   elsif( $iter_from eq "first" ) {
-      $iter_from = ITER_FIRST;
-   }
-   elsif( $iter_from eq "last" ) {
-      $iter_from = ITER_LAST;
-   }
-   else {
-      croak "Unrecognised 'iter_from' value %s";
-   }
-
-   my $on_iter;
-   if( defined $iter_from ) {
-      $on_iter = delete $args{on_iter};
-      ref $on_iter eq "CODE" or croak "Expected 'on_iter' to be a CODE ref";
-   }
-
-   my $on_watched = $args{on_watched};
-
-   my $pdef = $self->can_property( $property )
-      or croak "Class ".$self->classname." does not have a property $property";
+   my ( $pdef, %args ) = @_;
 
    my $callbacks = {};
    my $on_updated = delete $args{on_updated};
@@ -811,34 +575,52 @@ sub watch_property
          or croak "Expected '$name' as a CODE ref";
    }
 
+   return $callbacks;
+}
+
+sub watch_property              { shift->_watch_property( shift, 0, @_ ) }
+sub watch_property_with_initial { shift->_watch_property( shift, 1, @_ ) }
+
+sub _watch_property
+{
+   my $self = shift;
+   my ( $property, $want_initial, %args ) = @_;
+
+   # Detect void-context legacy uses
+   defined wantarray or
+      croak "->watch_property in void context no longer useful - it now returns a Future";
+
+   my $pdef = $self->can_property( $property )
+      or croak "Class ".$self->classname." does not have a property $property";
+
+   my $callbacks = _watchcbs_from_args( $pdef, %args );
+
    # Smashed properties behave differently
    my $smash = $pdef->smashed;
 
    if( my $cbs = $self->{props}->{$property}->{cbs} ) {
       if( $want_initial and !$smash ) {
-         $self->get_property(
-            property => $property,
-            on_value => sub {
+         return $self->get_property( $property )
+            ->then( sub {
                $callbacks->{on_set} and $callbacks->{on_set}->( $_[0] );
                $callbacks->{on_updated} and $callbacks->{on_updated}->( $_[0] );
                push @$cbs, $callbacks;
-               $on_watched->() if $on_watched;
-            },
-         );
+               Future->done;
+            });
       }
       elsif( $want_initial and $smash ) {
          my $cache = $self->{props}->{$property}->{cache};
          $callbacks->{on_set} and $callbacks->{on_set}->( $cache );
          $callbacks->{on_updated} and $callbacks->{on_updated}->( $cache );
          push @$cbs, $callbacks;
-         $on_watched->() if $on_watched;
+         return Future->done;
       }
       else {
          push @$cbs, $callbacks;
-         $on_watched->() if $on_watched;
+         return Future->done;
       }
 
-      return;
+      die "UNREACHED";
    }
 
    $self->{props}->{$property}->{cbs} = [ $callbacks ];
@@ -849,56 +631,110 @@ sub watch_property
          $callbacks->{on_set} and $callbacks->{on_set}->( $cache );
          $callbacks->{on_updated} and $callbacks->{on_updated}->( $cache );
       }
-      $on_watched->() if $on_watched;
-      return;
+
+      return Future->done;
    }
 
-   my $conn = $self->{conn};
-   my $request;
-   if( $iter_from ) {
-      $conn->_ver_can_iter or croak "Server is too old to support MSG_WATCH_ITER";
-      $pdef->dimension == DIM_QUEUE or croak "Can only iterate on queue-dimension properties";
+   my $client = $self->{client};
 
-      $request = Tangence::Message->new( $conn, MSG_WATCH_ITER )
-         ->pack_int( $self->id )
-         ->pack_str( $property )
-         ->pack_int( $iter_from );
-   }
-   else {
-      $request = Tangence::Message->new( $conn, MSG_WATCH )
+   my $request = Tangence::Message->new( $client, MSG_WATCH )
          ->pack_int( $self->id )
          ->pack_str( $property )
          ->pack_bool( $want_initial );
+
+   $client->request(
+      request => $request,
+   )->then( sub {
+      my ( $message ) = @_;
+      my $code = $message->code;
+
+      if( $code == MSG_WATCHING ) {
+         Future->done;
+      }
+      else {
+         Future->fail( "Unexpected response code $code", tangence => );
+      }
+   });
+}
+
+=head2 watch_property_with_iter
+
+   ( $iter, $first_idx, $last_idx ) =
+      $proxy->watch_property_with_iter( $property, $iter_from, %callbacks )->get
+
+A variant of C<watch_property> that installs a watch on the given property of
+the server object, and additionally returns an iterator object that can be
+used to lazily fetch the values stored in it.
+
+The C<$iter_from> value indicates which end of the queue the iterator should
+start from; C<ITER_FIRST> to start at index 0, or C<ITER_LAST> to start at the
+highest-numbered index. The iterator is created atomically with installing the
+watch.
+
+=cut
+
+sub watch_property_with_iter
+{
+   my $self = shift;
+   my ( $property, $iter_from, %args ) = @_;
+
+   # Detect void-context legacy uses
+   defined wantarray or
+      croak "->watch_property_with_iter in void context no longer useful - it now returns a Future";
+
+   if( $iter_from eq "first" ) {
+      $iter_from = ITER_FIRST;
+   }
+   elsif( $iter_from eq "last" ) {
+      $iter_from = ITER_LAST;
+   }
+   else {
+      croak "Unrecognised 'iter_from' value %s";
    }
 
-   $conn->request(
-      request => $request,
+   my $pdef = $self->can_property( $property )
+      or croak "Class ".$self->classname." does not have a property $property";
 
-      on_response => sub {
-         my ( $message ) = @_;
-         my $code = $message->code;
+   my $callbacks = _watchcbs_from_args( $pdef, %args );
 
-         if( $code == MSG_WATCHING ) {
-            $on_watched->() if $on_watched;
-         }
-         elsif( $code == MSG_WATCHING_ITER ) {
-            $on_watched->() if $on_watched;
-            my $iter_id = $message->unpack_int();
-            my $first_idx = $message->unpack_int();
-            my $last_idx  = $message->unpack_int();
+   # Smashed properties behave differently
+   my $smashed = $pdef->smashed;
 
-            my $iter = Tangence::ObjectProxy::_PropertyIterator->new( $self, $iter_id, $pdef->type );
-            $on_iter->( $iter, $first_idx, $last_idx );
-         }
-         elsif( $code == MSG_ERROR ) {
-            my $msg = $message->unpack_str();
-            $on_error->( $msg );
-         }
-         else {
-            $on_error->( "Unexpected response code $code" );
-         }
-      },
-   );
+   if( my $cbs = $self->{props}->{$property}->{cbs} ) {
+      die "TODO: need to synthesize a second iterator";
+   }
+
+   $self->{props}->{$property}->{cbs} = [ $callbacks ];
+
+   if( $smashed ) {
+      die "TODO: need to synthesize an iterator";
+   }
+
+   my $client = $self->{client};
+   $client->_ver_can_iter or croak "Server is too old to support MSG_WATCH_ITER";
+   $pdef->dimension == DIM_QUEUE or croak "Can only iterate on queue-dimension properties";
+
+   $client->request(
+      request => Tangence::Message->new( $client, MSG_WATCH_ITER )
+         ->pack_int( $self->id )
+         ->pack_str( $property )
+         ->pack_int( $iter_from ),
+   )->then( sub {
+      my ( $message ) = @_;
+      my $code = $message->code;
+
+      if( $code == MSG_WATCHING_ITER ) {
+         my $iter_id = $message->unpack_int();
+         my $first_idx = $message->unpack_int();
+         my $last_idx  = $message->unpack_int();
+
+         my $iter = Tangence::ObjectProxy::_PropertyIterator->new( $self, $iter_id, $pdef->type );
+         Future->done( $iter, $first_idx, $last_idx );
+      }
+      else {
+         Future->fail( "Unexpected response code $code", tangence => );
+      }
+   });
 }
 
 sub handle_request_UPDATE
@@ -1067,29 +903,19 @@ sub _update_property_objset
    }
 }
 
-=head2 $proxy->unwatch_property( %args )
+=head2 unwatch_property
+
+   $proxy->unwatch_property( $property )
 
 Removes a property watches on the given property on the server object that was
 previously installed using C<watch_property>.
 
-Takes the following named arguments:
-
-=over 8
-
-=item property => STRING
-
-Name of the property
-
-=back
-
 =cut
 
 sub unwatch_property
 {
    my $self = shift;
-   my %args = @_;
-
-   my $property = delete $args{property} or croak "Need a property";
+   my ( $property ) = @_;
 
    $self->can_property( $property )
       or croak "Class ".$self->classname." does not have a property $property";
@@ -1097,9 +923,9 @@ sub unwatch_property
    # TODO: mark iterators as destroyed and invalid
    delete $self->{props}->{$property};
 
-   my $conn = $self->{conn};
-   $conn->request(
-      request => Tangence::Message->new( $conn, MSG_UNWATCH )
+   my $client = $self->{client};
+   $client->request(
+      request => Tangence::Message->new( $client, MSG_UNWATCH )
          ->pack_int( $self->id )
          ->pack_str( $property ),
 
@@ -1125,105 +951,90 @@ sub new
    return bless [ @_ ], $class;
 }
 
-sub obj { shift->[0] }
-sub id  { shift->[1] }
-sub conn { shift->obj->{conn} }
+sub obj    { shift->[0] }
+sub id     { shift->[1] }
+sub client { shift->obj->{client} }
 
 sub DESTROY
 {
    my $self = shift;
 
-   return unless $self->obj and my $id = $self->id and my $conn = $self->conn;
+   return unless $self->obj and my $id = $self->id and my $client = $self->client;
 
-   $conn->request(
-      request => Tangence::Message->new( $conn, MSG_ITER_DESTROY )
+   $client->request(
+      request => Tangence::Message->new( $client, MSG_ITER_DESTROY )
          ->pack_int( $id ),
 
       on_response => sub {},
    );
 }
 
-=head2 $iter->next_forward( %args )
+=head2 next_forward
 
-=head2 $iter->next_backward( %args )
+   ( $index, @more ) = $iter->next_forward( $count )->get
+
+=head2 next_backward
+
+   ( $index, @more ) = $iter->next_backward( $count )->get
 
 Requests the next items from the iterator. C<next_forward> moves forwards
 towards higher-numbered indices, and C<next_backward> moves backwards towards
-lower-numbered indices.
+lower-numbered indices. If C<$count> is unspecified, a default of 1 will
+apply.
 
-The following arguments are recognised:
-
-=over 8
-
-=item count => INT
-
-Optional. Gives the number of elements requested. Will default to 1 if not
-provided.
-
-=item on_more => CODE
-
-Callback to invoke when the new elements are returned. This will be invoked
-with the index of the first element returned, and the new elements. Note that
-there may be fewer elements returned than were requested, if the end of the
-queue was reached. Specifically, there will be no new elements if the iterator
-is already at the end.
-
- $on_more->( $index, @items )
-
-=back
+The returned future wil yield the index of the first element returned, and the
+new elements. Note that there may be fewer elements returned than were
+requested, if the end of the queue was reached. Specifically, there will be no
+new elements if the iterator is already at the end.
 
 =cut
 
 sub next_forward
 {
    my $self = shift;
-   $self->_next( direction => ITER_FWD, @_ );
+   $self->_next( ITER_FWD, @_ );
 }
 
 sub next_backward
 {
    my $self = shift;
-   $self->_next( direction => ITER_BACK, @_ );
+   $self->_next( ITER_BACK, @_ );
 }
 
 sub _next
 {
    my $self = shift;
-   my %args = @_;
+   my ( $direction, $count ) = @_;
+
+   # Detect void-context legacy uses
+   defined wantarray or
+      croak "->next_forward/backward in void context no longer useful - it now returns a Future";
 
    my $obj = $self->obj;
    my $id  = $self->id;
    my $element_type = $self->[2];
 
-   my $on_more  = $args{on_more} or croak "Expected 'on_more' as a CODE ref";
-   my $on_error = $args{on_error} || $obj->{on_error};
+   my $client = $self->client;
 
-   my $conn = $self->conn;
-   $conn->request(
-      request => Tangence::Message->new( $conn, MSG_ITER_NEXT )
+   $client->request(
+      request => Tangence::Message->new( $client, MSG_ITER_NEXT )
          ->pack_int( $id )
-         ->pack_int( $args{direction} )
-         ->pack_int( $args{count} || 1 ),
+         ->pack_int( $direction )
+         ->pack_int( $count || 1 ),
+   )->then( sub {
+      my ( $message ) = @_;
+      my $code = $message->code;
 
-      on_response => sub {
-         my ( $message ) = @_;
-         my $code = $message->code;
-
-         if( $code == MSG_ITER_RESULT ) {
-            $on_more->(
-               $message->unpack_int(),
-               $message->unpack_all_sametype( $element_type ),
-            );
-         }
-         elsif( $code == MSG_ERROR ) {
-            my $msg = $message->unpack_str();
-            $on_error->( $msg );
-         }
-         else {
-            $on_error->( "Unexpected response code $code" );
-         }
+      if( $code == MSG_ITER_RESULT ) {
+         Future->done(
+            $message->unpack_int(),
+            $message->unpack_all_sametype( $element_type ),
+         );
       }
-   );
+      else {
+         Future->fail( "Unexpected response code $code", tangence => );
+      }
+   });
 }
 
 =head1 AUTHOR
